/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2018 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.40                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit.c for LPC1768 and LPC1766
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"
#ifdef __ICCARM__  // IAR
  #include "exceptions.h"         /* CMSIS compatible irq interface */
#endif

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS 72000000uL
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS)   /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART  (OS_FSYS/4) /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000u)
#endif

#ifndef   OS_USE_VARINTTABLE        /* The interrupt vector table   */
  #define OS_USE_VARINTTABLE (0)    /* may be located in RAM        */
#endif

#define OS_Systick SysTick_Handler  /* Map OS systick handler to CMSIS compatible handler name */

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifdef OS_LIBMODE_SAFE
  #undef    OS_VIEW_IFSELECT
  #define   OS_VIEW_IFSELECT  OS_VIEW_DISABLED  // Communication not supported in OS_LIBMODE_SAFE
#else
  #ifndef   OS_VIEW_IFSELECT
    #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
  #endif
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*       Currently UART0 is supported and the standard
*       setup enables UART0 per default.
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (0)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400)
  #endif
#endif

/*********************************************************************
*
*       Vector table
*/
#if (defined __SES_ARM)           // SEGGER Embedded Studio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __CROSSWORKS_ARM)  // Rowley CrossStudio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __ICCARM__)        // IAR
  #define __Vectors    __vector_table
#elif (defined __GNUC__)          // GCC
  extern unsigned char __Vectors;
#elif (defined __CC_ARM)          // KEIL
  extern unsigned char __Vectors;
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/
#define NVIC_VTOR         (*(volatile OS_U32*) (0xE000ED08uL))

#define SYS_INT_CTRL_STATE       (*(volatile OS_U32*)(0xE000ED04uL))
#define SYS_PENDSTSET            26

#define SYSPRI1_ADDR             (0xE000ED18uL)
#define SYSHND_CTRL_ADDR         (0xE000ED24uL)  // System Handler Control and State
#define SYSHND_CTRL              (*(volatile OS_U32*) (SYSHND_CTRL_ADDR))

#define NVIC_SYS_HND_CTRL_MEM    (0x00010000uL)  // Mem manage fault enable
#define NVIC_SYS_HND_CTRL_BUS    (0x00020000uL)  // Bus fault enable
#define NVIC_SYS_HND_CTRL_USAGE  (0x00040000uL)  // Usage fault enable

#define NVIC_PRIOBASE_ADDR       (0xE000E400uL)
#define NVIC_ENABLE_ADDR         (0xE000E100uL)
#define NVIC_DISABLE_ADDR        (0xE000E180uL)
#define NVIC_VTOREG_ADDR         (0xE000ED08uL)

#define NUM_INTERRUPTS           (16+35)

#define PINCON_BASE_ADDR         (0x4002C000uL)
#define PINSEL0                  (*(volatile unsigned long *)(PINCON_BASE_ADDR + 0x00))
#define PINSEL1                  (*(volatile unsigned long *)(PINCON_BASE_ADDR + 0x04))

#define PCONP                    (*(volatile unsigned long *)(0x400FC0C4uL))

/*****  PLL sfrs ****************************************************/
#define FLASHCFG                 (*(volatile OS_U32*)(0x400FC000uL))

#define SYSTEM_CONTROL_BLOCK_BASE     (0x400FC000uL)
#define SCS                           (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x01A0))

#define CCLKCFG                       (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x0104))
#define USBCLKCFG                     (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x0108))
#define CLKSRCSEL                     (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x010C))
#define PCLKSEL0                      (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x01A8))
#define PCLKSEL1                      (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x01AC))

#define PLL0CON                       (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x0080))
#define PLL0CFG                       (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x0084))
#define PLL0STAT                      (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x0088))
#define PLL0FEED                      (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x008C))

#define PLL1CON                       (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x00A0))
#define PLL1CFG                       (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x00A4))
#define PLL1STAT                      (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x00A8))
#define PLL1FEED                      (*(volatile OS_U32*)(SYSTEM_CONTROL_BLOCK_BASE + 0x00AC))

/*****  PLL values **************************************************/
#define SCS_OSCRANGE_BIT               (4)
#define SCS_OSCEN_BIT                  (5)
#define SCS_OSCSTAT_BIT                (6)

#define PLL0STAT_PLOCK_BIT            (26)
#define PLL1STAT_PLOCK_BIT            (10)

#define PLL0CFG_NSEL_BIT              (16)
#define PLL1CFG_PSEL_BIT               (5)

#define PLL0_MValue                   (11)
#define PLL0_NValue                    (0)
#define CCLKDivValue                   (3)
#define USBCLK_DIV_VALUE               (5)
#define PLL1_MValue                    (3)
#define PLL1_PValue                    (0)

#define SYS_TICK_BASE_ADDR       (0xE000E010uL)
#define SYS_TICK_CONTROL         (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x00))
#define SYS_TICK_RELOAD          (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x04))
#define SYS_TICK_VALUE           (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x08))
#define SYS_TICK_CALIBRATION     (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x0C))
#define SYS_TICK_ENABLE_BIT      (0)
#define SYS_TICK_INT_ENABLE_BIT  (1)
#define SYS_TICK_CLK_SOURCE_BIT  (2)

#define OS_TIMER_RELOAD           (OS_PCLK_TIMER / OS_TICK_FREQ - 1)
#if (OS_TIMER_RELOAD >= 0x100000000uL)
  #error "Systick can not be used, please check configuration"
#endif

/*****  Interrupt ID numbers ****************************************/
#define ISR_ID_NMI               (2)                // NMI fault
#define ISR_ID_HARD              (3)                // Hard fault
#define ISR_ID_MPU               (4)                // MPU fault
#define ISR_ID_BUS               (5)                // Bus fault
#define ISR_ID_USAGE             (6)                // Usage fault
#define ISR_ID_SVCALL            (11)               // SVCall
#define ISR_ID_DEBUG             (12)               // Debug monitor
#define ISR_ID_PENDSV            (14)               // PendSV
#define ISR_ID_SYSTICK           (15)               // System Tick
#define ISR_ID_USART1            (53)               // USART1
#define ISR_ID_USART2            (54)               // USART2

#define OS_ISR_ID_TICK           ISR_ID_SYSTICK     // We use Sys-Timer

/****** UART sfdr definition ****************************************/
#define UART0_BASE_ADDR   0x4000C000uL

#define UART_RBR_OFFS     0x00
#define UART_THR_OFFS     0x00
#define UART_IER_OFFS     0x04
#define UART_IIR_OFFS     0x08
#define UART_FCR_OFFS     0x08
#define UART_LCR_OFFS     0x0C
#define UART_LSR_OFFS     0x14
#define UART_SCR_OFFS     0x1C
#define UART_DLL_OFFS     0x00
#define UART_DLM_OFFS     0x04
#define UART_FDR_OFFS     0x28

/****** Assign UART sfrs used for embOSView communication ***********/
#define OS_UART_RBR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_RBR_OFFS))
#define OS_UART_THR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_THR_OFFS))
#define OS_UART_IER   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_IER_OFFS))
#define OS_UART_IIR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_IIR_OFFS))
#define OS_UART_FCR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_FCR_OFFS))
#define OS_UART_LCR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_LCR_OFFS))
#define OS_UART_LSR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_LSR_OFFS))
#define OS_UART_SCR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_SCR_OFFS))
#define OS_UART_DLL   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_DLL_OFFS))
#define OS_UART_DLM   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_DLM_OFFS))
#define OS_UART_FDR   (*(volatile OS_U8*)(UART0_BASE_ADDR + UART_FDR_OFFS))

#define RX_FULL_INT_ENABLE_BIT  0
#define TX_EMPTY_INT_ENABLE_BIT 1
#define RX_ERROR_INT_ENABLE_BIT 2
#define INT_PENDING_BIT         0

#define UART_INT_MASK         0x0Eu
#define UART_ERROR_INT_STATUS 0x06u
#define UART_RX_INT_STATUS    0x04u
#define UART_TX_INT_STATUS    0x02u

#define OS_UART_BASE_ADDR       (UART0_BASE_ADDR)
#define OS_UART_INT_INDEX       (21)
#define UART_PCONP_BIT          (3)
#define OS_UART_PINSEL_RX       PINSEL0
#define OS_UART_PINSEL_TX       PINSEL0
#define PINSEL_UART_MODE_RX_POS (6)
#define PINSEL_UART_MODE_TX_POS (4)
#define PINSEL_UART_MODE_RX_VAL (1)
#define PINSEL_UART_MODE_TX_VAL (1)

#define OS_COM_ISR_Handler      UART0_IRQHandler

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKMEM.h"
  const OS_U32 OS_JLINKMEM_BufferSize = 32u;  // Size of the communication buffer for JLINKMEM
#else
  const OS_U32 OS_JLINKMEM_BufferSize = 0;    // Communication not used
#endif

#if OS_USE_VARINTTABLE
  //
  // The interrupt vector table may be located anywhere in RAM
  //
  #ifdef __ICCARM__  // IAR
    #pragma data_alignment=512
    __no_init void (*g_pfnRAMVectors[NUM_INTERRUPTS]) (void);
  #endif  // __ICCARM__

  #ifdef __CC_ARM    // KEIL
     __attribute__ (zero_init, aligned(512)) void (*g_pfnRAMVectors[NUM_INTERRUPTS])(void);
  #endif

  #ifdef __GNUC__    // GCC
    void (*g_pfnRAMVectors[NUM_INTERRUPTS]) (void) __attribute__ ((aligned (512)));
  #endif
#else
  #define g_pfnRAMVectors  NULL
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return SYS_TICK_VALUE;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return SYS_INT_CTRL_STATE & (1uL << SYS_PENDSTSET);
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_Systick()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    _SysTick exception. SysTick is used as OS timer tick.
*
*  NOTES:
*    (1) It has to be inserted in the interrupt vector table, if RAM
*        vectors are not used. Therefore it is declared public
*/
#ifdef __cplusplus
extern "C" {
#endif
void OS_Systick(void);
#ifdef __cplusplus
}
#endif
void OS_Systick(void) {
  OS_EnterNestableInterrupt();
  OS_TICK_Handle();
  #if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  OS_U8 TickPrio;

  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 0, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_IncDI();
  SCS = (1uL << SCS_OSCEN_BIT);                           // Enable main OSC
  while ((SCS & (1uL << SCS_OSCSTAT_BIT)) == 0);          // Wait for Oscillator to be ready
  CLKSRCSEL = 0x1;                                        // Select main OSC, 12MHz, as the PLL clock source
  PLL0CFG   = PLL0_MValue | (PLL0_NValue << PLL0CFG_NSEL_BIT);
  PLL0CON   = 1;                                          // Enable PLL, disconnected
  PLL0FEED  = 0xAA;
  PLL0FEED  = 0x55;
  while (((PLL0STAT & (1uL << PLL0STAT_PLOCK_BIT)) == 0));// Check lock bit status
  CCLKCFG   = CCLKDivValue;                               // Set clock divider
  PLL0CON   = 3;                                          // Enable and connect
  PLL0FEED  = 0xAA;
  PLL0FEED  = 0x55;

  PLL1CFG   = PLL1_MValue | (PLL1_PValue << PLL1CFG_PSEL_BIT);
  PLL1CON   = 1;                                          // Enable PLL, disconnected
  PLL1FEED  = 0xAA;
  PLL1FEED  = 0x55;
  while (((PLL1STAT & (1uL << PLL1STAT_PLOCK_BIT)) == 0));// Check lock bit status
  PLL1CON   = 3;                                          // Enable and connect
  PLL1FEED  = 0xAA;
  PLL1FEED  = 0x55;

  FLASHCFG  = 0x0000303A;                                 // Flash accesses use 4 CPU clocks. For up to 80 MHz CPU clock.

  //
  // Initialize OS timer, clock soure = core clock
  //
  SYS_TICK_CONTROL &= ~(1uL << SYS_TICK_ENABLE_BIT);      // Stop counting
  SYS_TICK_CONTROL |=  (1uL << SYS_TICK_CLK_SOURCE_BIT);  // Select core clock = 100MHz
  SYS_TICK_CONTROL |=  (1uL << SYS_TICK_INT_ENABLE_BIT);  // Enable system tick timer interrupt
  SYS_TICK_VALUE    = 0;                                  // Reset counter
  // Set timer period
  SYS_TICK_RELOAD   = OS_TIMER_RELOAD;
  SYS_TICK_CONTROL |= (1uL << SYS_TICK_ENABLE_BIT);       // Enable counting
  //
  // Install Systick Timer Handler and enable timer interrupt
  //
  OS_ARM_InstallISRHandler(OS_ISR_ID_TICK, (OS_ISR_HANDLER*)OS_Systick);
  //
  // Initialize NVIC vector base address. Might be necessary for RAM targets or application not running from 0
  //
  NVIC_VTOR = (OS_U32)&__Vectors;
  //
  // Set the interrupt priority for the system timer to 2nd lowest level to ensure the timer can preempt PendSV handler
  //
  OS_ARM_ISRInit(OS_USE_VARINTTABLE, NUM_INTERRUPTS, (void*)__Vectors, g_pfnRAMVectors);
  OS_ARM_ISRSetPrio(OS_ISR_ID_TICK, 0xFFu);              // Set to lowest level, ALL BITS set
  TickPrio  = OS_ARM_ISRSetPrio(OS_ISR_ID_TICK, 0xFFu);  // Read priority back to examine relevant preemption-level-bits
  TickPrio -= 1;                                         // Set to low preemption level, 1 level higher than lowest
  OS_ARM_ISRSetPrio(OS_ISR_ID_TICK, TickPrio);
  OS_ARM_EnableISR(OS_ISR_ID_TICK);
  //
  // Setup values for usec precise system time functions
  //
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Configure and initialize SEGGER SystemView
  //
#if OS_PROFILE
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {     // Idle loop: No task is ready to execute
  while (1) {
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))     // Enter CPU halt mode when not in DEBUG build and J-Link communication not used
      #ifdef __ICCARM__  // IAR
        #if (__VER__ < 6300000)
          __asm(" wfi");
        #else
          __asm volatile(" wfi");
        #endif
      #endif
      #ifdef __CC_ARM    // KEIL
        __wfi();
      #endif
      #ifdef __GNUC__    // GCC
        __asm(" wfi");
      #endif
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
  Cnt  = OS_TIMER_RELOAD - SYS_TICK_VALUE;
  //
  // Check if timer interrupt pending ...
  //
  if (SYS_INT_CTRL_STATE & (1uL << SYS_PENDSTSET)) {
    Cnt = OS_TIMER_RELOAD - SYS_TICK_VALUE;  // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / 1000u));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)

static void   _JLINKMEM_OnRx       (OS_U8 Data);
static void   _JLINKMEM_OnTx       (void);
static OS_INT _JLINKMEM_GetNextChar(void);

/*********************************************************************
*
*       _JLINKMEM_OnRx()
*
*  Function description
*    This function is called when the next character is received
*/
static void _JLINKMEM_OnRx(OS_U8 Data) {
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _JLINKMEM_OnTx()
*
*  Function description
*    This function is called when the next character was sent
*/
static void _JLINKMEM_OnTx(void) {
  OS_OnTx();
}

/*********************************************************************
*
*       _JLINKMEM_GetNextChar()
*
*  Function description
*    Gets the next character
*/
static OS_INT _JLINKMEM_GetNextChar(void) {
  return OS_COM_GetNextChar();
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via J-Link
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKMEM_SendChar(c);
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes J-Link communication for embOSView
*/
void OS_COM_Init(void) {
  JLINKMEM_SetpfOnRx(_JLINKMEM_OnRx);
  JLINKMEM_SetpfOnTx(_JLINKMEM_OnTx);
  JLINKMEM_SetpfGetNextChar(_JLINKMEM_GetNextChar);
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

#define BAUDDIVIDE ((OS_PCLK_UART+OS_BAUDRATE*8L)/(OS_BAUDRATE*16L))

/*********************************************************************
*
*       OS_COM_ISR_Handler() embOS USART interrupt handler
*       handles both, Rx and Tx interrupt
*/
#if (defined __GNUC__)
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wunused-but-set-variable" // Avoid variable 'Dummy' set but not used warning
#endif
void OS_COM_ISR_Handler(void);
void OS_COM_ISR_Handler(void) {
  OS_U32          UartStatus;
  volatile OS_U32 Dummy;
  OS_EnterNestableInterrupt();
  UartStatus = OS_UART_IIR;                                         // Examine interrupt identification register
  if ((UartStatus & UART_INT_MASK) == UART_ERROR_INT_STATUS) {      // Error pending ?
    Dummy = OS_UART_RBR;                                            // Discard data
    Dummy = OS_UART_LSR;                                            // reset error
  } else if ((UartStatus & UART_INT_MASK) == UART_RX_INT_STATUS) {  // Data received?
    OS_OnRx(OS_UART_RBR);                                           // Process actual byte
  } else if ((UartStatus & UART_INT_MASK) == UART_TX_INT_STATUS) {  // Tx interrupt pending?
    if (OS_OnTx() != 0) {                                           // No more character sent?
      OS_UART_IER &= ~(1uL << TX_EMPTY_INT_ENABLE_BIT);             // Disable Tx interrupts
    }
  }
  OS_LeaveNestableInterrupt();
}
#if (defined __GNUC__)
  #pragma GCC diagnostic pop
#endif

/*********************************************************************
*
*       OS_COM_Send1()
*       Never call this function directly from your application
*/
void OS_COM_Send1(OS_U8 c) {
  OS_UART_THR  = c;                                // Send character
  OS_UART_IER |= (1uL << TX_EMPTY_INT_ENABLE_BIT); // Enable Tx interrupts
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize UART for OSView
*/
void OS_COM_Init(void) {
  OS_U8 Priority;
  // Setup Port-Mode to alternate function
  OS_UART_PINSEL_RX &= ~(0x03uL << PINSEL_UART_MODE_RX_POS);
  OS_UART_PINSEL_TX &= ~(0x03uL << PINSEL_UART_MODE_TX_POS);
  OS_UART_PINSEL_RX |=  (PINSEL_UART_MODE_RX_VAL << PINSEL_UART_MODE_RX_POS);
  OS_UART_PINSEL_TX |=  (PINSEL_UART_MODE_TX_VAL << PINSEL_UART_MODE_TX_POS);

  PCONP       |= (1uL << UART_PCONP_BIT);    // Enable UART unit
  OS_UART_IER  = 0x00;                       // Initially disable all interrupts
  OS_UART_LCR  = 0x80;                       // Set DLAB to initialize Baudrate generator
  OS_UART_DLL  = (BAUDDIVIDE & 0xFF);
  OS_UART_DLM  = ((BAUDDIVIDE >> 8) & 0xFF);
  OS_UART_LCR &= ~0x80;                      // reset DLAB to lock baudrate generator access
  OS_UART_LCR  = 0x03                        // 8 data bits
               | (0 << 2)                    // 1 stop bit
               | (0 << 3)                    // NO parity
               | (0 << 4)                    // Parity setting (bit 5:4) does not care
               | (0 << 6)                    // Disable Break transmission
               | (0 << 7);                   // Clear DLAB
  OS_UART_FCR  = (1uL << 0)                  // Enable FIFO (according to documentation)
               | (1uL << 1)                  // Reset RxFIFO
               | (1uL << 2)                  // Reset TxFIFO
               | (0 << 6)                    // Set trigger-level for RxFIFO
               ;
  /*  Install OS UART interrupt handler */
  OS_ARM_InstallISRHandler(OS_UART_INT_INDEX, &OS_COM_ISR_Handler);  // OS UART interrupt vector.
  OS_ARM_EnableISR(OS_UART_INT_INDEX);                               // Enable UART interrupt vector.
  Priority  = OS_ARM_ISRSetPrio(OS_UART_INT_INDEX, 0xFFu);           // Read back priority to examine preemption level bits
  Priority -= 1;                                                     // Set priority one level above lowest priority
  OS_ARM_ISRSetPrio(OS_UART_INT_INDEX, Priority);
  OS_UART_IER |= (1uL << RX_FULL_INT_ENABLE_BIT);                    // Enable Rx interrupts
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

#ifndef OS_LIBMODE_SAFE  // Communication not supported in OS_LIBMODE_SAFE
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}
#endif

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/****** End Of File *************************************************/
