/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2018 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.40                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_nRF51822.c for nRF51822 CPU and BBC micro:bit eval board.
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"
#include "nrf.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency   */
  #define OS_FSYS SystemCoreClock
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer */
  #define OS_PCLK_TIMER  (OS_FSYS)  /* May vary from CPU clock    */
#endif                              /* depending on CPU           */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART  */
  #define OS_PCLK_UART   (OS_FSYS)  /* May vary from CPU clock    */
#endif                              /* depending on CPU           */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ   (1000u)
#endif

#define OS_TIMER_RELOAD  (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (-1)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400)
  #endif
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       Vector table
*/
#if (defined __SES_ARM)           // SEGGER Embedded Studio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __CROSSWORKS_ARM)  // Rowley CrossStudio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __ICCARM__)        // IAR
  #define __Vectors    __vector_table
#elif (defined __TI_ARM__)        // TI CodeComposer
  extern unsigned char interruptVectors;
  #define __Vectors    interruptVectors
#elif (defined __GNUC__)          // GCC
  extern unsigned char __Vectors;
#elif (defined __CC_ARM)          // KEIL
  extern unsigned char __Vectors;
#elif (defined (__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050))
  extern unsigned char __Vectors;  // ARMCC Clang
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/
#define NVIC_VTOR         (*(volatile OS_U32*) (0xE000ED08uL))

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKMEM.h"
  const OS_U32 OS_JLINKMEM_BufferSize = 32u;  // Size of the communication buffer for JLINKMEM
#else
  const OS_U32 OS_JLINKMEM_BufferSize = 0;    // Communication not used
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  NRF_TIMER0->TASKS_CAPTURE[3] = 1;
  return NRF_TIMER0->CC[3];
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return NRF_TIMER0->EVENTS_COMPARE[0];
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       TIMER0_IRQHandler()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    _SysTick exception. SysTick is used as OS timer tick.
*/
#ifdef __cplusplus
extern "C" {
#endif
void TIMER0_IRQHandler(void);
#ifdef __cplusplus
}
#endif
#if OS_PROFILE
unsigned int SEGGER_SYSVIEW_TickCnt = 0;
#endif
void TIMER0_IRQHandler(void) {
  unsigned long old_ccompare;
  unsigned long diff = 0;

  OS_EnterNestableInterrupt();
  NRF_TIMER0->EVENTS_COMPARE[0]  = 0;
  do {
    old_ccompare                 = NRF_TIMER0->CC[0];
    NRF_TIMER0->CC[0]            = old_ccompare + OS_TIMER_RELOAD;
    NRF_TIMER0->TASKS_CAPTURE[1] = 1;
    diff                         = NRF_TIMER0->CC[1] - old_ccompare;
    OS_TICK_Handle();
#if OS_PROFILE
    SEGGER_SYSVIEW_TickCnt++;
#endif
  } while (diff >= OS_TIMER_RELOAD);

  #if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 1, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_IncDI();
  //
  // Initialize HFCLK XTAL
  //
  NRF_CLOCK->TASKS_HFCLKSTART = 1;
  //
  // Initialize timer
  //
  NVIC_EnableIRQ(TIMER0_IRQn);
  NRF_TIMER0->TASKS_SHUTDOWN = 1;
  NRF_TIMER0->INTENCLR       = TIMER_INTENCLR_COMPARE1_Msk
                             | TIMER_INTENCLR_COMPARE2_Msk
                             | TIMER_INTENCLR_COMPARE3_Msk;
  NRF_TIMER0->INTENSET       = TIMER_INTENSET_COMPARE0_Msk;
  NRF_TIMER0->MODE           = TIMER_MODE_MODE_Timer;
  NRF_TIMER0->BITMODE        = TIMER_BITMODE_BITMODE_32Bit;
  NRF_TIMER0->PRESCALER      = 0x0;
  NRF_TIMER0->CC[0]          = OS_TIMER_RELOAD;
  NRF_TIMER0->TASKS_CLEAR    = 1;
  NRF_TIMER0->TASKS_START    = 1;
  //
  // Initialize NVIC vector base address. Might be necessary for RAM targets or application not running from 0
  //
  NVIC_VTOR = (OS_U32)&__Vectors;
  //
  // Set the interrupt priority for the system timer to 2nd lowest level to ensure the timer can preempt PendSV handler
  //
  NVIC_SetPriority(TIMER0_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);
  //
  // Setup values for usec precise system time functions
  //
  SysTimerConfig.TimerFreq = SystemCoreClock;
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Configure and initialize SEGGER SystemView
  //
#if OS_PROFILE
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
  Cnt  = _OS_GetHWTimerCycles() % OS_TIMER_RELOAD;
  //
  // Check if timer interrupt pending ...
  //
  if (_OS_GetHWTimer_IntPending() != 0) {
    Cnt = _OS_GetHWTimerCycles() % OS_TIMER_RELOAD;  // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / 1000u));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)

static void   _JLINKMEM_OnRx       (OS_U8 Data);
static void   _JLINKMEM_OnTx       (void);
static OS_INT _JLINKMEM_GetNextChar(void);

/*********************************************************************
*
*       _JLINKMEM_OnRx()
*
*  Function description
*    This function is called when the next character is received
*/
static void _JLINKMEM_OnRx(OS_U8 Data) {
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _JLINKMEM_OnTx()
*
*  Function description
*    This function is called when the next character was sent
*/
static void _JLINKMEM_OnTx(void) {
  OS_OnTx();
}

/*********************************************************************
*
*       _JLINKMEM_GetNextChar()
*
*  Function description
*    Gets the next character
*/
static OS_INT _JLINKMEM_GetNextChar(void) {
  return OS_COM_GetNextChar();
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via J-Link
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKMEM_SendChar(c);
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes J-Link communication for embOSView
*/
void OS_COM_Init(void) {
  JLINKMEM_SetpfOnRx(_JLINKMEM_OnRx);
  JLINKMEM_SetpfOnTx(_JLINKMEM_OnTx);
  JLINKMEM_SetpfGetNextChar(_JLINKMEM_GetNextChar);
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

#define OS_BAUDDIVIDE ((OS_PCLK_UART + (OS_BAUDRATE * 16u)) / (OS_BAUDRATE * 32u)) - 1u

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via UART
*/
void OS_COM_Send1(OS_U8 c) {
}

/*********************************************************************
*
*       _OS_ISR_Rx()
*
*  Function description
*    Interrupt handler for UART reception
*/
static void _OS_ISR_Rx (void) {
}

/*********************************************************************
*
*       _OS_ISR_Tx()
*
*  Function description
*    Interrupt handler for UART transmission
*/
static void _OS_ISR_Tx (void) {
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes Uart communication for embOSView
*/
void OS_COM_Init(void) {
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via UDP
*/
void OS_COM_Send1(OS_U8 c) {
  UDP_Process_Send1(c);
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes UDP communication for embOSView
*/
void OS_COM_Init(void) {
  UDP_Process_Init();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Dummy routine.
*/
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/****** End Of File *************************************************/
