/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2018 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.40                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_STM32F051R8_CMSIS.c for ST STM32F051R8
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"
#include "stm32f0xx.h"  // Device specific header file, contains CMSIS

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS SystemCoreClock
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS)   /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART  (OS_FSYS)   /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000u)
#endif

#define OS_TIMER_RELOAD (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*       Currently the standard code enables OS_UART 1 per default.
*       UART is configured to be used on IAR STM32F051R8-SK.
*       OS_UART=1: RS232 on STM32F051R8-SK (USART1)
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (1)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400)
  #endif
#endif

/*********************************************************************
*
*       Vector table
*/
#if (defined __SES_ARM)           // SEGGER Embedded Studio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __CROSSWORKS_ARM)  // Rowley CrossStudio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __ICCARM__)        // IAR
  #define __Vectors    __vector_table
#elif (defined __GNUC__)          // GCC
  extern unsigned char __Vectors;
#elif (defined __CC_ARM)          // KEIL
  extern unsigned char __Vectors;
#elif defined (__CLANG__) || (defined (__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)) // ARM CLANG
  extern unsigned char __Vectors;
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKMEM.h"
  const OS_U32 OS_JLINKMEM_BufferSize = 32u;  // Size of the communication buffer for JLINKMEM
#else
  const OS_U32 OS_JLINKMEM_BufferSize = 0;    // Communication not used
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return SysTick->VAL;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return SCB->ICSR & SCB_ICSR_PENDSTSET_Msk;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    _SysTick exception. SysTick is used as OS timer tick.
*/
#ifdef __cplusplus
extern "C" {
#endif
void SysTick_Handler(void);      // Avoid warning, Systick_Handler is not prototyped in any CMSIS header
#ifdef __cplusplus
}
#endif
unsigned int SEGGER_SYSVIEW_TickCnt;
void SysTick_Handler(void) {
#if OS_PROFILE
  SEGGER_SYSVIEW_TickCnt++;
#endif
  OS_EnterNestableInterrupt();
  OS_TICK_Handle();
  #if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  // SysTimerConfig.TimerFreq will be set later, thus it is initialized with zero.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {0, OS_TICK_FREQ, 0, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_IncDI();
  //
  // We assume, the PLL and core clock was already set by the SystemInit() function
  // which was called from the startup code
  // Therefore, we don't have to initialize any hardware here,
  // we just ensure that the system clock variable is updated and then
  // set the periodic system timer tick for embOS.
  //
  SystemCoreClockUpdate();                // Update the system clock variable (might not have been set before)
  if (SysTick_Config(OS_TIMER_RELOAD)) {  // Setup SysTick Timer for 1 msec interrupts
    while (1);                            // Handle Error
  }
  //
  // Set the interrupt priority for the system timer to 2nd lowest level to ensure the timer can preempt PendSV handler
  //
  NVIC_SetPriority(SysTick_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);
  //
  // Setup values for usec precise system time functions
  //
  SysTimerConfig.TimerFreq = SystemCoreClock;
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Configure and initialize SEGGER SystemView
  //
#if OS_PROFILE
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {     // Idle loop: No task is ready to execute
  while (1) {            // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
      __WFI();           // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
  Cnt  = OS_TIMER_RELOAD - SysTick->VAL;
  //
  // Check if timer interrupt pending ...
  //
  if (SCB->ICSR & SCB_ICSR_PENDSTSET_Msk) {
    Cnt = OS_TIMER_RELOAD - SysTick->VAL;  // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / 1000u));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)

static void   _JLINKMEM_OnRx       (OS_U8 Data);
static void   _JLINKMEM_OnTx       (void);
static OS_INT _JLINKMEM_GetNextChar(void);

/*********************************************************************
*
*       _JLINKMEM_OnRx()
*
*  Function description
*    This function is called when the next character is received
*/
static void _JLINKMEM_OnRx(OS_U8 Data) {
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _JLINKMEM_OnTx()
*
*  Function description
*    This function is called when the next character was sent
*/
static void _JLINKMEM_OnTx(void) {
  OS_OnTx();
}

/*********************************************************************
*
*       _JLINKMEM_GetNextChar()
*
*  Function description
*    Gets the next character
*/
static OS_INT _JLINKMEM_GetNextChar(void) {
  return OS_COM_GetNextChar();
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via J-Link
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKMEM_SendChar(c);
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes J-Link communication for embOSView
*/
void OS_COM_Init(void) {
  JLINKMEM_SetpfOnRx(_JLINKMEM_OnRx);
  JLINKMEM_SetpfOnTx(_JLINKMEM_OnTx);
  JLINKMEM_SetpfGetNextChar(_JLINKMEM_GetNextChar);
}


#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

#define ISR_ID_USART          USART1_IRQn                      // USART3 ID = 27
#define USART_BASE_ADDR       (0x40013800uL)                   // USART 1 base address
#define OS_BRR_VALUE          (OS_PCLK_UART / OS_BAUDRATE)     // USART1 uses PCKL1 = HCKL / 4

#define USART_CR1            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x00))
#define USART_ISR            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x1C))
#define USART_RDR            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x24))
#define USART_TDR            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x28))
#define USART_BRR            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x0C))

#define RCC_BASE_ADDR        ((OS_U32)0x40021000uL)
#define RCC_AHBENR           (*(volatile OS_U32*)(RCC_BASE_ADDR + 0x14))
#define RCC_APB2ENR          (*(volatile OS_U32*)(RCC_BASE_ADDR + 0x18))

#define GPIOA_BASE_ADDR      ((OS_U32)0x48000000uL)
#define GPIOA_MODER          (*(volatile OS_U32*)(GPIOA_BASE_ADDR + 0x00))
#define GPIOA_OTYPER         (*(volatile OS_U32*)(GPIOA_BASE_ADDR + 0x04))
#define GPIOA_OSPEEDR        (*(volatile OS_U32*)(GPIOA_BASE_ADDR + 0x08))
#define GPIOA_AF_HIGH        (*(volatile OS_U32*)(GPIOA_BASE_ADDR + 0x24))

#define US_RXRDY             (0x20u)   // RXNE
#define USART_RX_ERROR_FLAGS (0x0Fu)   // ORE/NE/FE/PE
#define US_TXEMPTY           (0x80u)   // TXE

/*********************************************************************
*
*       USART1_IRQHandler
*
* Function description
*   The communication interrupt handler for UART communication
*   to embOSView.
*/
#if (defined __GNUC__)
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wunused-but-set-variable" // Avoid variable 'Dummy' set but not used warning
#endif
void USART1_IRQHandler(void);
void USART1_IRQHandler(void) {
  volatile OS_U32 Dummy;
  int UsartStatus;
  OS_EnterNestableInterrupt();
  UsartStatus = USART_ISR;                       // Examine status register
  do {
    if (UsartStatus & US_RXRDY) {                // Data received?
      if (UsartStatus & USART_RX_ERROR_FLAGS) {  // Any error ?
        Dummy = USART_RDR;                        // => Discard data
      } else {
        OS_OnRx(USART_RDR);                       // Process actual byte
      }
    }
    UsartStatus = USART_ISR;                     // Examine current status
  } while (UsartStatus & US_RXRDY);
  if (UsartStatus & US_TXEMPTY) {
    if (OS_OnTx()) {                             // No more characters to send ?
      USART_CR1 &= ~0x40uL;                      // Disable further tx interrupts
    }
  }
  OS_LeaveNestableInterrupt();
}
#if (defined __GNUC__)
  #pragma GCC diagnostic pop
#endif

/*********************************************************************
*
*       OS_COM_Send1()
*
*       Sends one character via UART. Never call this from your application
*/
void OS_COM_Send1(OS_U8 c) {
  USART_TDR  = (c & (OS_U16)0x01FF);
  USART_CR1 |= 0x40;                           // Enable tx interrupt
}

/*********************************************************************
*
*       OS_COM_Init()
*
*       Initialize the selected UART
*/
void OS_COM_Init(void) {
  //
  // Initialize clock and port for USART, selection specific
  //
  OS_IncDI();
  RCC_AHBENR    |= (1uL << 17);              // GPIOB CLK enable
  RCC_APB2ENR   |= (1uL << 14);              // Enable USART1 clock
                                             // GPIOC set alternate function for USART3
  GPIOA_AF_HIGH =  (1uL << 4)                // - Set pin_09 to AF1
                |  (1uL << 8);               // - Set pin_10 to AF1
                                             // GPIOC speed setting
  GPIOA_OSPEEDR =  (3uL << 18)               // - Pin_09
                |  (3uL << 20);              // - Pin_10
  GPIOA_OTYPER  =  0x0;                      // Output type: push-pull for pin_09 and pin_10
                                             // GPIOC alternate function mode
  GPIOA_MODER   |= (2uL << 18)               // - Pin_09 AF
                |  (2uL << 20);              // - Pin_10 AF
  //
  // USART3 settings
  //
  USART_CR1 = (1uL << 3)                     // Transmitter enable
            | (1uL << 2)                     // Receiver enable
            | (1uL << 5);                    // RX interrupt enable
  USART_BRR = OS_BRR_VALUE;                  // Set baudrate
  USART_CR1 |= (1uL << 0);                   // Enable USART
  NVIC_SetPriority(ISR_ID_USART, (1u << __NVIC_PRIO_BITS) - 4u);
  NVIC_EnableIRQ(ISR_ID_USART);
  OS_DecRI();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/****** End Of File *************************************************/
