/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2018 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.40                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit.c
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS 72000000uL
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS)   /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART  (OS_FSYS/2) /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000u)
#endif

#ifndef   OS_USE_VARINTTABLE        /* The interrupt vector table   */
  #define OS_USE_VARINTTABLE (0)    /* may be located in RAM        */
#endif

#define OS_Systick SysTick_Handler  /* Map OS systick handler to CMSIS compatible handler name */

#define OS_TIMER_RELOAD (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to OSView, please define it to be -1
*       Currently the standard code enables UART 0 per default
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (1)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400)
  #endif
#endif
/****** End of configurable options *********************************/

/*********************************************************************
*
*       Vector table
*/
#if (defined __SES_ARM)           // SEGGER Embedded Studio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __CROSSWORKS_ARM)  // Rowley CrossStudio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __ICCARM__)        // IAR
  #define __Vectors    __vector_table
#elif (defined __GNUC__)          // GCC
  extern unsigned char __Vectors;
#elif (defined __CC_ARM)          // KEIL
  extern unsigned char __Vectors;
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/
#define NVIC_VTOR         (*(volatile OS_U32*) (0xE000ED08uL))

#define SYS_TICK_BASE_ADDR       (0xE000E010uL)
#define SYS_TICK_CONTROL         (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x00))
#define SYS_TICK_RELOAD          (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x04))
#define SYS_TICK_VALUE           (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x08))
#define SYS_TICK_CALIBRATION     (*(volatile OS_U32*)(SYS_TICK_BASE_ADDR + 0x0C))
#define SYS_TICK_ENABLE_BIT      (0)
#define SYS_TICK_INT_ENABLE_BIT  (1)
#define SYS_TICK_CLK_SOURCE_BIT  (2)

#define SYS_INT_CTRL_STATE       (*(volatile OS_U32*)(0xE000ED04uL))
#define SYS_PENDSTSET            26

#define SYSCTRL_BASE_ADDR        (0x400FE000uL)
#define SYSCTRL_RCGC1            (*(volatile OS_U32*) (SYSCTRL_BASE_ADDR + 0x104))
#define SYSCTRL_RCGC2            (*(volatile OS_U32*) (SYSCTRL_BASE_ADDR + 0x108))
#define SYSCTRL_RCC              (*(volatile OS_U32*) (SYSCTRL_BASE_ADDR + 0x60))  // Run-mode clock config register
#define SYSCTRL_MISC             (*(volatile OS_U32*) (SYSCTRL_BASE_ADDR + 0x58))  // Interrupt status register
#define SYSCTRL_RIS              (*(volatile OS_U32*) (SYSCTRL_BASE_ADDR + 0x50))  // Raw interrupt status register

#define SYSCTRL_RCC_BYPASS       (0x00000800uL)  // PLL bypass
#define SYSCTRL_RCC_USE_SYSDIV   (0x00400000uL)  // Use sytem clock divider
#define SYSCTRL_RCC_IOSCDIS      (0x00000002uL)  // Internal oscillator disable
#define SYSCTRL_RCC_MOSCDIS      (0x00000001uL)  // Main oscillator disable
#define SYSCTRL_RCC_XTAL_MASK    (0x000003C0uL)  // Crystal attached to main osc
#define SYSCTRL_RCC_OSCSRC_MASK  (0x00000030uL)  // Oscillator input select
#define SYSCTRL_RCC_PWRDN        (0x00002000uL)  // PLL power down
#define SYSCTRL_RCC_OE           (0x00001000uL)  // PLL output enable
#define SYSCTRL_INT_PLL_LOCK     (0x00000040uL)  // PLL lock interrupt
#define SYSCTRL_RCC_SYSDIV_MASK  (0x07800000uL)  // System clock divider

#define SYSPRI1_ADDR             (0xE000ED18uL)
#define SYSHND_CTRL_ADDR         (0xE000ED24uL)  // System Handler Control and State
#define SYSHND_CTRL              (*(volatile OS_U32*) (SYSHND_CTRL_ADDR))

#define NVIC_SYS_HND_CTRL_MEM    (0x00010000uL)  // Mem manage fault enable
#define NVIC_SYS_HND_CTRL_BUS    (0x00020000uL)  // Bus fault enable
#define NVIC_SYS_HND_CTRL_USAGE  (0x00040000uL)  // Usage fault enable

#define NVIC_BASE_ADDR           (0xE000E000uL)

#define NVIC_PRIOBASE_ADDR       (0xE000E400uL)
#define NVIC_ENABLE_ADDR         (0xE000E100uL)
#define NVIC_DISABLE_ADDR        (0xE000E180uL)
#define NVIC_VTOREG_ADDR         (0xE000ED08uL)

#define NUM_INTERRUPTS           (16+68)

/*********************************************************************
*
* The following can be used as as arguments for the PLL activation
* if required in __low_level_init()
*
**********************************************************************
*/
#define SYSCTRL_SYSDIV_1         (0x07800000uL)  // Processor clock is osc/pll /1
#define SYSCTRL_SYSDIV_4         (0x01C00000uL)  // Processor clock is osc/pll /4
#define SYSCTRL_SYSDIV_10        (0x04C00000uL)  // Processor clock is osc/pll /10
#define SYSCTRL_USE_PLL          (0x00000000uL)  // System clock is the PLL clock
#define SYSCTRL_USE_OSC          (0x00003800uL)  // System clock is the osc clock
#define SYSCTRL_XTAL_6MHZ        (0x000002C0uL)  // External crystal is 6MHz
#define SYSCTRL_XTAL_8MHZ        (0x00000380uL)  // External crystal is 8MHz
#define SYSCTRL_OSC_MAIN         (0x00000000uL)  // Oscillator source is main osc

#define OS_TIMER_RELOAD           (OS_PCLK_TIMER / OS_TICK_FREQ)
#if (OS_TIMER_RELOAD >= 0x100000000uL)
  #error "Systick can not be used, please check configuration"
#endif

/*****  Interrupt ID numbers ****************************************/
#define ISR_ID_NMI               (2)                // NMI fault
#define ISR_ID_HARD              (3)                // Hard fault
#define ISR_ID_MPU               (4)                // MPU fault
#define ISR_ID_BUS               (5)                // Bus fault
#define ISR_ID_USAGE             (6)                // Usage fault
#define ISR_ID_SVCALL            (11)               // SVCall
#define ISR_ID_DEBUG             (12)               // Debug monitor
#define ISR_ID_PENDSV            (14)               // PendSV
#define ISR_ID_SYSTICK           (15)               // System Tick
#define ISR_ID_USART1            (53)               // USART1
#define ISR_ID_USART2            (54)               // USART2

#define OS_ISR_ID_TICK           ISR_ID_SYSTICK     // We use Sys-Timer
#define OS_ISR_TICK_PRIO         (0xFFu)            // Lowest priority

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKMEM.h"
  const OS_U32 OS_JLINKMEM_BufferSize = 32u;  // Size of the communication buffer for JLINKMEM
#else
  const OS_U32 OS_JLINKMEM_BufferSize = 0;    // Communication not used
#endif

#if OS_USE_VARINTTABLE
  //
  // The interrupt vector table may be located anywhere in RAM
  //
  #ifdef __ICCARM__  // IAR
    #pragma data_alignment=512
    __no_init void (*g_pfnRAMVectors[NUM_INTERRUPTS]) (void);
  #endif  // __ICCARM__

  #ifdef __CC_ARM    // KEIL
     __attribute__ (zero_init, aligned(512)) void (*g_pfnRAMVectors[NUM_INTERRUPTS])(void);
  #endif

  #ifdef __GNUC__    // GCC
    void (*g_pfnRAMVectors[NUM_INTERRUPTS]) (void) __attribute__ ((aligned (512)));
  #endif
#else
  #define g_pfnRAMVectors  NULL
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return SYS_TICK_VALUE;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return SYS_INT_CTRL_STATE & (1uL << SYS_PENDSTSET);
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_Systick()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    _SysTick exception. SysTick is used as OS timer tick.
*
* NOTES:
*   (1) It has to be inserted in the interrupt vector table, if RAM
*       vectors are not used. Therefore it is declared public
*/
#ifdef __cplusplus
extern "C" {
#endif
void OS_Systick(void);
#ifdef __cplusplus
}
#endif
void OS_Systick(void) {
  OS_EnterNestableInterrupt();
  OS_TICK_Handle();
  #if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used
*/
#define PERIPH_BASE                   ((OS_U32)0x40000000uL)
#define APB1PERIPH_BASE               PERIPH_BASE
#define APB2PERIPH_BASE               (PERIPH_BASE + 0x10000u)
#define AHBPERIPH_BASE                (PERIPH_BASE + 0x20000u)
#define RCC_BASE                      (AHBPERIPH_BASE + 0x1000u)
#define FLASH_BASE                    ((OS_U32)0x40022000uL)
#define RCC_CR                        (*(volatile OS_U32*)(RCC_BASE   + 0x00))
#define RCC_CFGR                      (*(volatile OS_U32*)(RCC_BASE   + 0x04))
#define RCC_CIR                       (*(volatile OS_U32*)(RCC_BASE   + 0x08))
#define RCC_APB2RSTR                  (*(volatile OS_U32*)(RCC_BASE   + 0x0C))
#define RCC_APB1RSTR                  (*(volatile OS_U32*)(RCC_BASE   + 0x10))
#define RCC_AHBENR                    (*(volatile OS_U32*)(RCC_BASE   + 0x14))
#define RCC_APB2ENR                   (*(volatile OS_U32*)(RCC_BASE   + 0x18))
#define RCC_APB1ENR                   (*(volatile OS_U32*)(RCC_BASE   + 0x1C))
#define RCC_BDCR                      (*(volatile OS_U32*)(RCC_BASE   + 0x20))
#define RCC_CSR                       (*(volatile OS_U32*)(RCC_BASE   + 0x24))
#define RCC_AHBRSTR                   (*(volatile OS_U32*)(RCC_BASE   + 0x28))
#define RCC_CFGR2                     (*(volatile OS_U32*)(RCC_BASE   + 0x2C))
#define FLASH_ACR                     (*(volatile OS_U32*)(FLASH_BASE + 0x00))

#define RCC_CR_HSION_BIT               (0)
#define RCC_CR_HSEON_BIT              (16)
#define RCC_CR_HSERDY_BIT             (17)
#define RCC_CR_PLL1ON_BIT             (24)
#define RCC_CR_PLL1RDY_BIT            (25)
#define RCC_CR_PLL2ON_BIT             (26)
#define RCC_CR_PLL2RDY_BIT            (27)
#define RCC_CR_PLL3ON_BIT             (28)
#define RCC_CR_PLL3RDY_BIT            (29)
#define RCC_CR_HSEON_RESET_MASK       ((OS_U32)0xFFFEFFFF)
#define RCC_CR_HSEBYP_RESET_MASK      ((OS_U32)0xFFFBFFFF)
#define RCC_CFGR_PLL1SRC_BIT          (16)
#define RCC_CFGR_HPRE_RESET_MASK      ((OS_U32)0xFFFFFF0F)
#define RCC_CFGR_PPRE1_RESET_MASK     ((OS_U32)0xFFFFF8FF)
#define RCC_CFGR_PPRE2_RESET_MASK     ((OS_U32)0xFFFFC7FF)
#define RCC_CFGR_ADCPRE_RESET_MASK    ((OS_U32)0xFFFF3FFF)
#define RCC_CFGR_HPRE_RESET_MASK      ((OS_U32)0xFFFFFF0F)
#define RCC_CFGR_SWS_RESET_MASK       ((OS_U32)0xFFFFFFFC)
#define RCC_CFGR_SW_RESET_MASK        ((OS_U32)0xFFFFFFF3)
#define RCC_CFGR_PLL1MUL_RESET_MASK   ((OS_U32)0xFFC0FFFF)
#define RCC_CRGR_PLL_SELECTED_MASK    ((OS_U32)0x0000000C)
#define ACR_LATENCY_MASK              ((OS_U32)0x00000038)
#define ACR_PRFTBE_MASK               ((OS_U32)0xFFFFFFEF)
#define FLASH_LATENCY_2_MASK          ((OS_U32)0x00000002)
#define FLASH_PREFETCH_BUFFER_ENABLE  ((OS_U32)0x00000010)
#define RCC_CFGR2_PREDIV1SRC_BIT      (16)
#define RCC_CFGR2_PREDIV1_RESET_MASK  ((OS_U32)0xFFFFFFF0)
#define RCC_CFGR2_PREDIV2_RESET_MASK  ((OS_U32)0xFFFFFF0F)
#define RCC_CFGR2_PLL2MUL_RESET_MASK  ((OS_U32)0xFFFFF0FF)

void OS_InitHW(void) {
  OS_U32   Val;
  volatile OS_U32 Dummy;
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 0, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_IncDI();
  //
  // Initialize NVIC vector base address. Might be necessary for RAM targets or application not running from 0
  //
  NVIC_VTOR = (OS_U32)&__Vectors;
  //
  // Set FLITF and SRAM Clock ON
  //
  RCC_AHBENR   = 0
               | (0 << 16)                   // ETHMACRXEN: Ethernet MAC RX clock enable. In the RMII mode, if this clock is enabled, the RMII clock of the MAC is also enabled.
               | (0 << 15)                   // ETHMACTXEN: Ethernet MAC TX clock enable. In the RMII mode, if this clock is enabled, the RMII clock of the MAC is also enabled.
               | (0 << 14)                   // ETHMACEN: Ethernet MAC clock enable.
               | (0 << 12)                   // OTGFSEN: USB OTG FS clock enable.
               | (0 <<  6)                   // CRCEN: CRC clock enable.
               | (1 <<  4)                   // FLITFEN: FLITF clock enable.
               | (1 <<  2)                   // SRAMEN: SRAM interface clock enable.
               | (0 <<  1)                   // DMA2EN: DMA2 clock enable.
               | (0 <<  0)                   // DMA1EN: DMA1 clock enable.
               ;
  RCC_APB2ENR  = 0x0;                        // Disable APB2 Peripheral clocks
  RCC_APB1ENR  = 0x0;                        // Disable APB1 Peripheral clocks
  //
  // Resets the RCC clock configuration to the default reset state.
  //
  RCC_CR       = (1 << RCC_CR_HSION_BIT);    // Set HSION bit
  RCC_CFGR    &= 0xF8FF0000uL;               // Reset SW[1:0], HPRE[3:0], PPRE1[2:0], PPRE2[2:0], ADCPRE[1:0] and MCO[2:0] bits
  RCC_CR      &= 0xFEF6FFFFuL;               // Reset HSEON[16], CSSON[19] and PLL1ON bits
  RCC_CR      &= 0xFFFBFFFFuL;               // Reset HSEBYP[18] bit
  RCC_CFGR    &= 0xFF80FFFFuL;               // Reset PLLSRC, PLLXTPRE, PLLMUL[3:0] and USBPRE bits
  RCC_CIR      = 0x0;                        // Disable all interrupts
  //
  // Configure HSE
  //
  RCC_CR      &= RCC_CR_HSEON_RESET_MASK;    // Reset HSEON[16]
  RCC_CR      &= RCC_CR_HSEBYP_RESET_MASK;   // Reset HSEBYP[18]
  RCC_CR      |= (1 << RCC_CR_HSEON_BIT);    // Enable HSE
  while (((RCC_CR & (1uL << RCC_CR_HSERDY_BIT)) == 0));   // Wait till HSE is ready (The RCC_CR_HSERDY_BIT (Bit 17) is 1 if the HSE is running.)
  //
  // Flash 2 wait state and Prefetch enable
  //
  FLASH_ACR   &= ACR_LATENCY_MASK;
  FLASH_ACR   |= FLASH_LATENCY_2_MASK;
  FLASH_ACR   &= ACR_PRFTBE_MASK;
  FLASH_ACR   |= FLASH_PREFETCH_BUFFER_ENABLE;
  //
  // HCLK = SYSCLK
  //
  Val      = RCC_CFGR;
  Val     &= RCC_CFGR_HPRE_RESET_MASK;       // 0xx: HCLK not divided
  RCC_CFGR = Val;
  //
  // PCLK2 = HCLK
  //
  Val      = RCC_CFGR;
  Val     &= RCC_CFGR_PPRE2_RESET_MASK;      // 0xx: HCLK not divided
  RCC_CFGR = Val;
  //
  // PCLK1 = HCLK/2
  //
  Val      = RCC_CFGR;
  Val     &= RCC_CFGR_PPRE1_RESET_MASK;
  Val     |= (1 << 10);                      // PPRE1[10:8]: b100: HCLK divided by 2
  RCC_CFGR = Val;
  //
  // PLL2 configuration: PLL2CLK = (HSE/5) * 8 = 40 MHz
  //
  Val       = RCC_CFGR2;
  Val      &= RCC_CFGR2_PREDIV2_RESET_MASK;
  Val      |= (1 << 6);                      // Set PREDIV2[7:4]: b0100: PREDIV2 input clock divided by 5
  RCC_CFGR2 = Val;
  Val       = RCC_CFGR2;
  Val      &= RCC_CFGR2_PLL2MUL_RESET_MASK;
  Val      |= (6 << 8);                      // Set PLL2MUL[11:8]: b0110: PLL2 clock entry x 8
  RCC_CFGR2 = Val;
  RCC_CR   |= (1 << RCC_CR_PLL2ON_BIT);         // Enable PLL2
  while (RCC_CR & (1 << RCC_CR_PLL2RDY_BIT)){}  // Wait till PLL2 is ready
  //
  // PLL1 configuration: PLL1CLK = (PLL2/5) * 9 = 72 MHz
  //
  Val       = RCC_CFGR2;
  Val      &= RCC_CFGR2_PREDIV1_RESET_MASK;
  Val      |= (1 << RCC_CFGR2_PREDIV1SRC_BIT);  // Select PLL2 as PREDIV1 clock entry
  Val      |= (1 << 2);                         // Set PREDIV1[3:0]: b0100: PREDIV1 input clock divided by 5
  RCC_CFGR2 = Val;
  Val       = RCC_CFGR;
  Val      &= RCC_CFGR_PLL1MUL_RESET_MASK;
  Val      |= ((1 << RCC_CFGR_PLL1SRC_BIT) | (7 << 18));  // Set PLL1SRC and PLL1MUL[21:18]: b0111: PLL1 input clock x 9
  RCC_CFGR  = Val;
  RCC_CR   |= (1 << RCC_CR_PLL1ON_BIT);         // Enable PLL1
  while (((RCC_CR & (1 << RCC_CR_PLL1RDY_BIT)) == 0));  // Wait until PLL1 is ready
  //
  // Add a short delay. Required!
  //
  for (Dummy = 0; Dummy < 1000; Dummy++) {
  }
  //
  // Select PLL1 as system clock source
  //
  Val       = RCC_CFGR;
  Val      &= RCC_CFGR_SW_RESET_MASK;
  Val      |= (1 << 1);                          // SW[1:0]: b10: PLL selected as system clock
  RCC_CFGR  = Val;
  while ((RCC_CFGR & RCC_CRGR_PLL_SELECTED_MASK) != 0x8) {} // Wait till PLL1 is used as system clock source
  //
  // Initialize OS timer, clock soure = core clock
  //
  SYS_TICK_RELOAD  = OS_TIMER_RELOAD;
  SYS_TICK_CONTROL = (1 << SYS_TICK_ENABLE_BIT) | (1 << SYS_TICK_CLK_SOURCE_BIT);
  //
  // Install Systick Timer Handler and enable timer interrupt
  //
  OS_ARM_ISRInit(OS_USE_VARINTTABLE, NUM_INTERRUPTS, (void*)__Vectors, g_pfnRAMVectors);
  OS_ARM_InstallISRHandler(OS_ISR_ID_TICK, (OS_ISR_HANDLER*)OS_Systick);
  OS_ARM_ISRSetPrio(OS_ISR_ID_TICK, OS_ISR_TICK_PRIO);
  OS_ARM_EnableISR(OS_ISR_ID_TICK);
  //
  // Setup values for usec precise system time functions
  //
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Configure and initialize SEGGER SystemView
  //
#if OS_PROFILE
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {     // Idle loop: No task is ready to execute
  while (1) {
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))     // Enter CPU halt mode when not in DEBUG build and J-Link communication not used
      #ifdef __ICCARM__  // IAR
        #if (__VER__ < 6300000)
          __asm(" wfi");
        #else
          __asm volatile(" wfi");
        #endif
      #endif
      #ifdef __CC_ARM    // KEIL
        __wfi();
      #endif
      #ifdef __GNUC__    // GCC
        __asm(" wfi");
      #endif
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
  Cnt  = (OS_TIMER_RELOAD) - SYS_TICK_VALUE;
  //
  // Check if timer interrupt pending ...
  //
  if (SYS_INT_CTRL_STATE & (1uL << SYS_PENDSTSET)) {
    Cnt = OS_TIMER_RELOAD - SYS_TICK_VALUE;  // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / 1000u));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)

static void   _JLINKMEM_OnRx       (OS_U8 Data);
static void   _JLINKMEM_OnTx       (void);
static OS_INT _JLINKMEM_GetNextChar(void);

/*********************************************************************
*
*       _JLINKMEM_OnRx()
*
*  Function description
*    This function is called when the next character is received
*/
static void _JLINKMEM_OnRx(OS_U8 Data) {
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _JLINKMEM_OnTx()
*
*  Function description
*    This function is called when the next character was sent
*/
static void _JLINKMEM_OnTx(void) {
  OS_OnTx();
}

/*********************************************************************
*
*       _JLINKMEM_GetNextChar()
*
*  Function description
*    Gets the next character
*/
static OS_INT _JLINKMEM_GetNextChar(void) {
  return OS_COM_GetNextChar();
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via J-Link
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKMEM_SendChar(c);
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes J-Link communication for embOSView
*/
void OS_COM_Init(void) {
  JLINKMEM_SetpfOnRx(_JLINKMEM_OnRx);
  JLINKMEM_SetpfOnTx(_JLINKMEM_OnTx);
  JLINKMEM_SetpfGetNextChar(_JLINKMEM_GetNextChar);
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

#define OS_COM_IsrHandler    USART2_IRQHandler
#define USART_BASE_ADDR      (0x40004400uL)
#define OS_BRR_VALUE         (OS_PCLK_UART / OS_BAUDRATE)
#define USART_SR             (*(volatile OS_U32*)(USART_BASE_ADDR + 0x00))
#define USART_DR             (*(volatile OS_U32*)(USART_BASE_ADDR + 0x04))
#define USART_BRR            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x08))
#define USART_CR1            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x0C))
#define USART_CR2            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x10))
#define USART_CR3            (*(volatile OS_U32*)(USART_BASE_ADDR + 0x14))
#define USART_GTPR           (*(volatile OS_U32*)(USART_BASE_ADDR + 0x18))
#define RCC_USART2EN         ((OS_U32)(1 << 17))

#define GPIOD_CRL            (*(volatile OS_U32*)(0x40011400uL))

#define US_RXRDY             (0x20)   // RXNE
#define USART_RX_ERROR_FLAGS (0x0F)   // ORE/NE/FE/PE

#define AFIO_BASE_ADDR       0x40010000uL
#define AFIO_MAPR            (*(volatile OS_U32*)(AFIO_BASE_ADDR + 0x04))

/*********************************************************************
*
*       OS_COM_IsrHandler
*
* Function description
*   The communication interrupt handler for UART communication
*   to embOSView.
*
* NOTES:
*   (1) It has to be inserted in the interrupt vector table, if RAM
*       vectors are not used. Therefore is is declared public
*/
#if (defined __GNUC__)
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wunused-but-set-variable" // Avoid variable 'Dummy' set but not used warning
#endif
void OS_COM_IsrHandler(void) {
  volatile OS_U32 Dummy;
  int UsartStatus;
  OS_EnterNestableInterrupt();
  UsartStatus = USART_SR;                        // Examine status register
  do {
    if (UsartStatus & US_RXRDY) {                // Data received?
      if (UsartStatus & USART_RX_ERROR_FLAGS) {  // Any error ?
        Dummy = USART_DR;                        // => Discard data
      } else {
        OS_OnRx(USART_DR);                       // Process actual byte
      }
    }
    else {
      if (OS_OnTx()) {                           // No more characters to send ?
        USART_CR1 &= ~0x40uL;                    // Disable further tx interrupts
      }
    }
    UsartStatus = USART_SR;                      // Examine current status
  }
  while (UsartStatus & US_RXRDY);
  OS_LeaveNestableInterrupt();
}
#if (defined __GNUC__)
  #pragma GCC diagnostic pop
#endif

/*********************************************************************
*
*       OS_COM_Send1()
*
*       Sends one character via UART. Never call this from your application
*/
void OS_COM_Send1(OS_U8 c) {
  while ((USART_SR & (1 << 7)) == 0);
  USART_DR   = (c & (OS_U16)0x01FF);
  USART_CR1 |= 0x40;                           // enable tx interrupt
}

/*********************************************************************
*
*       OS_COM_Init()
*
*       Initialize the selected UART
*/
void OS_COM_Init(void) {
  OS_U8 Priority;
  RCC_APB1ENR |= RCC_USART2EN;         // Enable USART2 clock
  RCC_APB2ENR |= (1 << 5) | (1 << 0);  // Enable GPIO port needed for USART, enable alternate function
  AFIO_MAPR   |= (1 << 3);             // Remap USART2
  GPIOD_CRL    = (GPIOD_CRL & 0xF00FFFFFuL) | 0x04B00000uL;

  USART_CR2   = 0x200;
  USART_CR1   = 0x2C;
  USART_CR3   = 0x00;
  USART_BRR   = OS_BRR_VALUE;
  USART_CR1  |= 0x2000;         // enable uart
  //
  // Install USART2 Handler with preemtion level one above lowest level to ensure communication
  //
  OS_ARM_InstallISRHandler(ISR_ID_USART2, (OS_ISR_HANDLER*)OS_COM_IsrHandler);
  OS_ARM_ISRSetPrio(ISR_ID_USART2, 0xFFu);              // Set lowest Priority, ALL BITS set
  Priority  = OS_ARM_ISRSetPrio(ISR_ID_USART2, 0xFFu);  // Read back priority to examine preemption level bits
  Priority -= 1;                                        // Set priority one level above lowest priority
  OS_ARM_ISRSetPrio(ISR_ID_USART2, Priority);
  OS_ARM_EnableISR(ISR_ID_USART2);
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/****** End Of File *************************************************/
