/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2018 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.40                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_STM32L1xx_CMSIS.c for STM32L1xx
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"
#include "stm32l1xx.h"  // Device specific header file, contains CMSIS

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration of tickless support
*/
#ifndef   OS_USE_TICKLESS_SUPPORT
  #define OS_USE_TICKLESS_SUPPORT (0)
#endif

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS SystemCoreClock
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #if (OS_USE_TICKLESS_SUPPORT == 0)
    #define OS_PCLK_TIMER (OS_FSYS)/* May vary from CPU clock       */
  #else
    #define OS_PCLK_TIMER (OS_FSYS/64)/* May vary from CPU clock    */
  #endif
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART (OS_FSYS)    /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000u)
#endif

#define OS_TIMER_RELOAD (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifdef OS_LIBMODE_SAFE
  #undef    OS_VIEW_IFSELECT
  #define   OS_VIEW_IFSELECT  OS_VIEW_DISABLED  // Communication not supported in OS_LIBMODE_SAFE
#else
  #ifndef   OS_VIEW_IFSELECT
    #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
  #endif
#endif

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*       Currently the standard code enables OS_UART 1 per default.
*       OS_UART=2: RS232_2 on IAR STM32LxB-SK (USART2)
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (2)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400)
  #endif
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       Vector table
*/
#if (defined __SES_ARM)           // SEGGER Embedded Studio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __CROSSWORKS_ARM)  // Rowley CrossStudio
  extern int _vectors;
  #define __Vectors    _vectors
#elif (defined __ICCARM__)        // IAR
  #define __Vectors    __vector_table
#elif (defined __GNUC__)          // GCC
  extern unsigned char __Vectors;
#elif (defined __CC_ARM)          // KEIL
  extern unsigned char __Vectors;
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/
#define NVIC_VTOR         (*(volatile OS_U32*) (0xE000ED08uL))

//
// Uart register definitions
//
#define USART2_BASE_ADDR     (0x40004400uL)
#define USART_SR             (*(volatile OS_U32*)(USART2_BASE_ADDR + 0x00))
#define USART_DR             (*(volatile OS_U32*)(USART2_BASE_ADDR + 0x04))
#define USART_BRR            (*(volatile OS_U32*)(USART2_BASE_ADDR + 0x08))
#define USART_CR1            (*(volatile OS_U32*)(USART2_BASE_ADDR + 0x0C))
#define USART_CR2            (*(volatile OS_U32*)(USART2_BASE_ADDR + 0x10))
#define USART_CR3            (*(volatile OS_U32*)(USART2_BASE_ADDR + 0x14))
#define US_RXRDY             (0x20u)   // RXNE
#define USART_RX_ERROR_FLAGS (0x0Fu)   // ORE/NE/FE/PE
#define US_TXEMPTY           (0x80u)   // TXE

#define GPIOD_BASE_ADDR      (0x40020C00uL)
#define GPIOD_MODER          (*(volatile OS_U32*)(GPIOD_BASE_ADDR + 0x00u))
#define GPIOD_ODR            (*(volatile OS_U32*)(GPIOD_BASE_ADDR + 0x14u))
#define GPIOD_AFRL           (*(volatile OS_U32*)(GPIOD_BASE_ADDR + 0x20u))
#define GPIOD_AFRH           (*(volatile OS_U32*)(GPIOD_BASE_ADDR + 0x24u))

#define RCC_BASE_ADDR        (0x40023800uL)
#define RCC_AHBENR           (*(volatile OS_U32*)(RCC_BASE_ADDR + 0x1C))
#define RCC_APB1ENR          (*(volatile OS_U32*)(RCC_BASE_ADDR + 0x24))

#define GPIOD_EN             (3u)
#define RS232_EN             (0u)
#define GPIO_AF7             (7u)
#define GPIO_MODER_MASK      (0x03u)
#define GPIOD0_MODER         (0u)
#define GPIOD5_MODER         (10u)
#define GPIOD6_MODER         (12u)
#define GPIO_MODER_OUTPUT    (1u)
#define GPIO_MODER_ALTFUNC   (2u)
#define GPIO_AFRL_MASK       (0x0Fu)
#define GPIOD5_AFRL          (20u)
#define GPIOD6_AFRL          (24u)

#define OS_BRR_VALUE         (OS_PCLK_UART / OS_BAUDRATE)

//
// TIMER2
//
#define TIM2_ADDR            0x40000000u
#define TIM2_CR1             (*(volatile OS_U16*)(TIM2_ADDR + 0x00u))
#define TIM2_DIER            (*(volatile OS_U16*)(TIM2_ADDR + 0x0Cu))
#define TIM2_SR              (*(volatile OS_U16*)(TIM2_ADDR + 0x10u))
#define TIM2_CNT             (*(volatile OS_U16*)(TIM2_ADDR + 0x24u))
#define TIM2_PSC             (*(volatile OS_U16*)(TIM2_ADDR + 0x28u))
#define TIM2_ARR             (*(volatile OS_U16*)(TIM2_ADDR + 0x2Cu))
#define RCC_TIM2EN           0u
#define OS_ISR_ID_TIM2       (28u + 16u)
#define TIM2_MAX_TICKS       (0xFFFF / OS_TIMER_RELOAD)
#define OS_ISR_ID_TICK       OS_ISR_ID_TIM2

#define OS_TIMER_RELOAD      (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKMEM.h"
  const OS_U32 OS_JLINKMEM_BufferSize = 32u;  // Size of the communication buffer for JLINKMEM
#else
  const OS_U32 OS_JLINKMEM_BufferSize = 0;    // Communication not used
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return SysTick->VAL;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return SCB->ICSR & SCB_ICSR_PENDSTSET_Msk;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

#if (OS_USE_TICKLESS_SUPPORT == 0)
/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    _SysTick exception. SysTick is used as OS timer tick.
*/
#ifdef __cplusplus
extern "C" {
#endif
void SysTick_Handler(void);      // Avoid warning, Systick_Handler is not prototyped in any CMSIS header
#ifdef __cplusplus
}
#endif
void SysTick_Handler(void) {
  OS_EnterNestableInterrupt();
  OS_TICK_Handle();
  #if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}
#else
/*********************************************************************
*
*       TIM2_IRQHandler()
*
* Function description
*   TIM2_IRQHandler is used as OS timer tick.
*/
void TIM2_IRQHandler(void);
void TIM2_IRQHandler(void) {
  TIM2_SR &= ~(1uL << 0u);
  OS_EnterNestableInterrupt();

  OS_TICK_Handle();
  #if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       _EndTicklessMode()
*
* Function description
*   This function calculates how long we actually stayed in sleep mode
*   and corrects the system time.
*   The hardware timer will be reset to default tick time.
*/
static void _EndTicklessMode(void) {
  OS_U16 NumTicks;
  OS_U16 Cnt;
  OS_U16 IReq;

  if (OS_Global.TicklessExpired) {
    // The timer interrupt was executed => we completetd the sleep time
    OS_AdjustTime(OS_Global.TicklessFactor);
  } else {
    Cnt  = TIM2_CNT;
    IReq = TIM2_SR & (1uL << 0u);
    if (IReq) {
      OS_AdjustTime(OS_Global.TicklessFactor);
    } else {
      //
      // We assume OS_TIMER_RELOAD Counts per tick and hardware timer which counts up.
      //
      NumTicks   = Cnt / OS_TIMER_RELOAD;
      Cnt       -= (NumTicks * OS_TIMER_RELOAD);
      TIM2_CNT = Cnt;
      OS_AdjustTime(NumTicks);
    }
  }
  TIM2_ARR = OS_TIMER_RELOAD; // Set default value for 1 tick
}
#endif

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  // SysTimerConfig.TimerFreq will be set later, thus it is initialized with zero.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {0, OS_TICK_FREQ, 0, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_IncDI();
  //
  // We assume, the PLL and core clock was already set by the SystemInit() function
  // which was called from the startup code
  // Therefore, we don't have to initialize any hardware here,
  // we just ensure that the system clock variable is updated and then
  // set the periodic system timer tick for embOS.
  //
  SystemCoreClockUpdate();                // Update the system clock variable (might not have been set before)
#if (OS_USE_TICKLESS_SUPPORT == 0)
  if (SysTick_Config(OS_TIMER_RELOAD)) {  // Setup SysTick Timer for 1 msec interrupts
    while (1);                            // Handle Error
  }
  //
  // Set the interrupt priority for the system timer to 2nd lowest level to ensure the timer can preempt PendSV handler
  //
  NVIC_SetPriority(SysTick_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);
#else
  RCC_APB1ENR |= (1u << RCC_TIM2EN);   // Enable Timer2 clock
  TIM2_PSC   = 63u;                    // Set prescaler to 63 => Divide by 64
  TIM2_ARR   = OS_TIMER_RELOAD;        // Set Auto reload value
  TIM2_DIER  = 0x01;                   // Update interrupt enabled
  TIM2_CR1   = 0x01;                   // Start timer, up counting
  NVIC_SetPriority(TIM2_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);
  NVIC_EnableIRQ(TIM2_IRQn);
#endif
  //
  // Initialize NVIC vector base address. Might be necessary for RAM targets or application not running from 0
  //
  NVIC_VTOR = (OS_U32)&__Vectors;
  //
  // Setup values for usec precise system time functions
  //
  SysTimerConfig.TimerFreq = SystemCoreClock;
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Configure and initialize SEGGER SystemView
  //
#if OS_PROFILE
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {     // Idle loop: No task is ready to execute
#if (OS_USE_TICKLESS_SUPPORT == 1)
  OS_TIME IdleTicks;

  OS_DI();
  IdleTicks = OS_GetNumIdleTicks();
  if (IdleTicks > 1) {
    //
    // Check what the hardware can do. Typically, the hardware time has a limit.
    // If we exceed that limit, simply go to what the hardware can do.
    // Example: If we can wait 50 Ticks (i = 50), but the hardware time can not do
    // more than 10, then we program it to use only 10 and wake up 5 times, every 10ms.
    //
    if ((OS_U32)IdleTicks > TIM2_MAX_TICKS) {
      IdleTicks = TIM2_MAX_TICKS;
    }
    OS_StartTicklessMode(IdleTicks, &_EndTicklessMode);
    TIM2_ARR = (OS_TIMER_RELOAD * IdleTicks) - TIM2_CNT; // Set compare register
  }
  OS_EI();
#endif
  while (1) {
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
      __WFI();           // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
#if (OS_USE_TICKLESS_SUPPORT == 0)
  Cnt  = OS_TIMER_RELOAD - SysTick->VAL;
  //
  // Check if timer interrupt pending ...
  //
  if (SCB->ICSR & SCB_ICSR_PENDSTSET_Msk) {
    Cnt = OS_TIMER_RELOAD - SysTick->VAL;  // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
#else
  Cnt = TIM2_CNT;
  //
  // Check if timer interrupt pending ...
  //
  if (TIM2_SR & (1uL << 0u)) {
    Cnt = TIM2_CNT;
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
#endif
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / 1000u));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)

static void   _JLINKMEM_OnRx       (OS_U8 Data);
static void   _JLINKMEM_OnTx       (void);
static OS_INT _JLINKMEM_GetNextChar(void);

/*********************************************************************
*
*       _JLINKMEM_OnRx()
*
*  Function description
*    This function is called when the next character is received
*/
static void _JLINKMEM_OnRx(OS_U8 Data) {
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _JLINKMEM_OnTx()
*
*  Function description
*    This function is called when the next character was sent
*/
static void _JLINKMEM_OnTx(void) {
  OS_OnTx();
}

/*********************************************************************
*
*       _JLINKMEM_GetNextChar()
*
*  Function description
*    Gets the next character
*/
static OS_INT _JLINKMEM_GetNextChar(void) {
  return OS_COM_GetNextChar();
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via J-Link
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKMEM_SendChar(c);
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes J-Link communication for embOSView
*/
void OS_COM_Init(void) {
  JLINKMEM_SetpfOnRx(_JLINKMEM_OnRx);
  JLINKMEM_SetpfOnTx(_JLINKMEM_OnTx);
  JLINKMEM_SetpfGetNextChar(_JLINKMEM_GetNextChar);
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

/*********************************************************************
*
*       USART2_IRQHandler()
*
* Function description
*   The communication interrupt handler for UART communication
*   to embOSView.
*/
#if (defined __GNUC__)
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wunused-but-set-variable" // Avoid variable 'Dummy' set but not used warning
#endif
void USART2_IRQHandler(void);
void USART2_IRQHandler(void) {
  volatile OS_U32 Dummy;
  int UsartStatus;
  OS_EnterNestableInterrupt();
  UsartStatus = USART_SR;                        // Examine status register
  do {
    if (UsartStatus & US_RXRDY) {                // Data received?
      if (UsartStatus & USART_RX_ERROR_FLAGS) {  // Any error ?
        Dummy = USART_DR;                        // => Discard data
      } else {
        OS_OnRx(USART_DR);                       // Process actual byte
      }
    }
    UsartStatus = USART_SR;                      // Examine current status
  } while (UsartStatus & US_RXRDY);
  if (UsartStatus & US_TXEMPTY) {
    if (OS_OnTx()) {                             // No more characters to send ?
      USART_CR1 &= ~0x40uL;                      // Disable further tx interrupts
    }
  }
  OS_LeaveNestableInterrupt();
}
#if (defined __GNUC__)
  #pragma GCC diagnostic pop
#endif

/*********************************************************************
*
*       OS_COM_Send1()
*
* Function description
*   Sends one character via UART. Never call this from your application
*/
void OS_COM_Send1(OS_U8 c) {
  USART_DR   = (c & (OS_U16)0x01FF);
  USART_CR1 |= 0x40;                           // Enable tx interrupt
}

/*********************************************************************
*
*       OS_COM_Init()
*
* Function description
*   Initialize the selected UART
*/
void OS_COM_Init(void) {
  //
  // Initialize USART
  //
  OS_IncDI();
  //
  // Set port pins for Uart2 (Tx: PD5, Rx: PD6)
  //
  RCC_AHBENR  |= (1u << GPIOD_EN);                                                                                                     // Enable GPIOD clock
  GPIOD_MODER &= ~((GPIO_MODER_MASK << GPIOD0_MODER) | (GPIO_MODER_MASK << GPIOD5_MODER) | (GPIO_MODER_MASK << GPIOD5_MODER));         // Clear Mode for Rx and Tx pin
  GPIOD_MODER |=  ((GPIO_MODER_OUTPUT << GPIOD0_MODER) | (GPIO_MODER_ALTFUNC << GPIOD5_MODER) | (GPIO_MODER_ALTFUNC << GPIOD6_MODER)); // Set alternate function mode for Rx and Tx pin and output for RS232_EN
  GPIOD_AFRL  &= ~((GPIO_AFRL_MASK << GPIOD5_AFRL) | (GPIO_AFRL_MASK << GPIOD6_AFRL));                                                 // Clear AFRL bits for Tx and Rx pin
  GPIOD_AFRL  |=  ((GPIO_AF7 << GPIOD5_AFRL) | (GPIO_AF7 << GPIOD6_AFRL));                                                             // Set AF7 for Rx and Tx pin
  GPIOD_ODR   &= ~(0x01u << RS232_EN);                                                                                                 // Clear port pin to power ST232CDR
  //
  // Initialize uart registers
  //
  RCC_APB1ENR |= (1uL << 17);   // Enable USART2 clock
  USART_CR2   = 0x200;
  USART_CR1   = 0x2C;
  USART_CR3   = 0x00;
  USART_BRR   = OS_BRR_VALUE;
  USART_CR1  |= 0x2000;         // Enable uart
  //
  // Install USART Handler with preemtion level one above lowest level to ensure communication during PendSV
  //
  NVIC_SetPriority(USART2_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);              // Set lowest Priority, ALL BITS set
  NVIC_EnableIRQ(USART2_IRQn);
  OS_DecRI();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

#ifndef OS_LIBMODE_SAFE  // Communication not supported in OS_LIBMODE_SAFE
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}
#endif

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/****** End Of File *************************************************/
