/*****************************************************************************
 *                   SEGGER Microcontroller GmbH & Co. KG                    *
 *            Solutions for real time microcontroller applications           *
 *****************************************************************************
 *                                                                           *
 *               (c) 2017 SEGGER Microcontroller GmbH & Co. KG               *
 *                                                                           *
 *           Internet: www.segger.com   Support: support@segger.com          *
 *                                                                           *
 *****************************************************************************/

/*****************************************************************************
 *                         Preprocessor Definitions                          *
 *                         ------------------------                          *
 * NO_FPU_ENABLE                                                             *
 *                                                                           *
 *   If defined, FPU will not be enabled.                                    *
 *                                                                           *
 * NO_STACK_INIT                                                             *
 *                                                                           *
 *   If defined, the stack pointer will not be initialised.                  *
 *                                                                           *
 * NO_SYSTEM_INIT                                                            *
 *                                                                           *
 *   If defined, the SystemInit() function will not be called. By default    *
 *   SystemInit() is called after reset to enable the clocks and memories to *
 *   be initialised prior to any C startup initialisation.                   *
 *                                                                           *
 * NO_VTOR_CONFIG                                                            *
 *                                                                           *
 *   If defined, the vector table offset register will not be configured.    *
 *                                                                           *
 * MEMORY_INIT                                                               *
 *                                                                           *
 *   If defined, the MemoryInit() function will be called. By default        *
 *   MemoryInit() is called after SystemInit() to enable an external memory  *
 *   controller.                                                             *
 *                                                                           *
 * STACK_INIT_VAL                                                            *
 *                                                                           *
 *   If defined, specifies the initial stack pointer value. If undefined,    *
 *   the stack pointer will be initialised to point to the end of the        *
 *   RAM segment.                                                            *
 *                                                                           *
 * VECTORS_IN_RAM                                                            *
 *                                                                           *
 *   If defined, the exception vectors will be copied from Flash to RAM.     *
 *                                                                           *
 *****************************************************************************/

  .syntax unified

  .global Reset_Handler
  .global _vectors

  .section .init, "ax"
  .thumb_func

  .equ VTOR_REG, 0xE000ED08
  .equ FPU_CPACR_REG, 0xE000ED88

#ifndef STACK_INIT_VAL
#define STACK_INIT_VAL __RAM_segment_end__
#endif

Reset_Handler:
#ifndef NO_STACK_INIT
  /* Initialise main stack */
  ldr r0, =STACK_INIT_VAL
  movs r1, #0x7
  bics r0, r1
  mov sp, r0
#endif

#ifndef NO_SYSTEM_INIT
  /* Initialise system */
  ldr r0, =SystemInit
  blx r0
#endif

#ifdef MEMORY_INIT
  ldr r0, =MemoryInit
  blx r0
#endif

#ifdef VECTORS_IN_RAM
  /* Copy exception vectors into RAM */
  ldr r0, =__vectors_start__
  ldr r1, =__vectors_end__
  ldr r2, =__vectors_ram_start__
1:
  cmp r0, r1
  beq 2f
  ldr r3, [r0]
  str r3, [r2]
  adds r0, r0, #4
  adds r2, r2, #4
  b 1b
2:
#endif

#ifndef NO_VTOR_CONFIG
  /* Configure vector table offset register */
  ldr r0, =VTOR_REG
#ifdef VECTORS_IN_RAM
  ldr r1, =_vectors_ram
#else
  ldr r1, =_vectors
#endif
  str r1, [r0]
#endif

#if (defined(__ARM_ARCH_FPV4_SP_D16__) || defined(__ARM_ARCH_FPV5_D16__)) && !defined(NO_FPU_ENABLE)
  /* Enable FPU */
  ldr r0, =FPU_CPACR_REG
  ldr r1, [r0]
  orr r1, r1, #(0xF << 20)
  str r1, [r0]
  dsb
  isb
#endif

  /* Jump to program start */
  b _start

/**
**===========================================================================
**  Program - Default_Handler
**  Abstract: This code gets called when the processor receives an
**    unexpected interrupt.
**===========================================================================
*/
    .section    .text.Default_Handler,"ax",%progbits
Default_Handler:
InfiniteLoop:
    b   InfiniteLoop
    .size   Default_Handler, .-Default_Handler

/**
**===========================================================================
**  Reset, Exception, and Interrupt vectors
**===========================================================================
*/

  .section .vectors, "ax"
  .code 16
  .align 0

  .extern __stack_end__

_vectors:
  .word       __stack_end__           /* Top of Stack */
  .word       Reset_Handler           /* Reset Handler */
  .word       NMI_Handler             /* NMI Handler */
  .word       HardFault_Handler       /* Hard Fault Handler */
  .word       MemManage_Handler       /* MPU Fault Handler */
  .word       BusFault_Handler        /* Bus Fault Handler */
  .word       UsageFault_Handler      /* Usage Fault Handler */
  .word       Default_Handler         /* Reserved */
  .word       Default_Handler         /* Reserved */
  .word       Default_Handler         /* Reserved */
  .word       Default_Handler         /* Reserved */
  .word       SVC_Handler             /* SVCall Handler */
  .word       DebugMon_Handler        /* Debug Monitor Handler */
  .word       Default_Handler         /* Reserved */
  .word       PendSV_Handler          /* PendSV Handler */
  .word       SysTick_Handler         /* SysTick Handler */

  /* External interrupts */

  .word       DMA_IRQHandler          /* 0 - DMA */
  .word       GPIO_EVEN_IRQHandler    /* 1 - GPIO_EVEN */
  .word       TIMER0_IRQHandler       /* 2 - TIMER0 */
  .word       USART0_RX_IRQHandler    /* 3 - USART0_RX */
  .word       USART0_TX_IRQHandler    /* 4 - USART0_TX */
  .word       USB_IRQHandler          /* 5 - USB */
  .word       ACMP0_IRQHandler        /* 6 - ACMP0 */
  .word       ADC0_IRQHandler         /* 7 - ADC0 */
  .word       DAC0_IRQHandler         /* 8 - DAC0 */
  .word       I2C0_IRQHandler         /* 9 - I2C0 */
  .word       I2C1_IRQHandler         /* 10 - I2C1 */
  .word       GPIO_ODD_IRQHandler     /* 11 - GPIO_ODD */
  .word       TIMER1_IRQHandler       /* 12 - TIMER1 */
  .word       TIMER2_IRQHandler       /* 13 - TIMER2 */
  .word       TIMER3_IRQHandler       /* 14 - TIMER3 */
  .word       USART1_RX_IRQHandler    /* 15 - USART1_RX */
  .word       USART1_TX_IRQHandler    /* 16 - USART1_TX */
  .word       LESENSE_IRQHandler      /* 17 - LESENSE */
  .word       USART2_RX_IRQHandler    /* 18 - USART2_RX */
  .word       USART2_TX_IRQHandler    /* 19 - USART2_TX */
  .word       UART0_RX_IRQHandler     /* 20 - UART0_RX */
  .word       UART0_TX_IRQHandler     /* 21 - UART0_TX */
  .word       UART1_RX_IRQHandler     /* 22 - UART1_RX */
  .word       UART1_TX_IRQHandler     /* 23 - UART1_TX */
  .word       LEUART0_IRQHandler      /* 24 - LEUART0 */
  .word       LEUART1_IRQHandler      /* 25 - LEUART1 */
  .word       LETIMER0_IRQHandler     /* 26 - LETIMER0 */
  .word       PCNT0_IRQHandler        /* 27 - PCNT0 */
  .word       PCNT1_IRQHandler        /* 28 - PCNT1 */
  .word       PCNT2_IRQHandler        /* 29 - PCNT2 */
  .word       RTC_IRQHandler          /* 30 - RTC */
  .word       BURTC_IRQHandler        /* 31 - BURTC */
  .word       CMU_IRQHandler          /* 32 - CMU */
  .word       VCMP_IRQHandler         /* 33 - VCMP */
  .word       LCD_IRQHandler          /* 34 - LCD */
  .word       MSC_IRQHandler          /* 35 - MSC */
  .word       AES_IRQHandler          /* 36 - AES */
  .word       EBI_IRQHandler          /* 37 - EBI */
  .word       EMU_IRQHandler          /* 38 - EMU */
_vectors_end:

#ifdef VECTORS_IN_RAM
  .section .vectors_ram, "ax"
  .align 0
  .global _vectors_ram

_vectors_ram:
  .space _vectors_end - _vectors, 0
#endif


/**
**===========================================================================
** Provide weak aliases for each Exception handler to the Default_Handler.
**===========================================================================
*/
  .macro  def_irq_handler handler_name
  .weak   \handler_name
  .set    \handler_name, Default_Handler
  .endm

  def_irq_handler     NMI_Handler
  def_irq_handler     HardFault_Handler
  def_irq_handler     MemManage_Handler
  def_irq_handler     BusFault_Handler
  def_irq_handler     UsageFault_Handler
  def_irq_handler     SVC_Handler
  def_irq_handler     DebugMon_Handler
  def_irq_handler     PendSV_Handler
  def_irq_handler     SysTick_Handler

  def_irq_handler     DMA_IRQHandler
  def_irq_handler     GPIO_EVEN_IRQHandler
  def_irq_handler     TIMER0_IRQHandler
  def_irq_handler     USART0_RX_IRQHandler
  def_irq_handler     USART0_TX_IRQHandler
  def_irq_handler     USB_IRQHandler
  def_irq_handler     ACMP0_IRQHandler
  def_irq_handler     ADC0_IRQHandler
  def_irq_handler     DAC0_IRQHandler
  def_irq_handler     I2C0_IRQHandler
  def_irq_handler     I2C1_IRQHandler
  def_irq_handler     GPIO_ODD_IRQHandler
  def_irq_handler     TIMER1_IRQHandler
  def_irq_handler     TIMER2_IRQHandler
  def_irq_handler     TIMER3_IRQHandler
  def_irq_handler     USART1_RX_IRQHandler
  def_irq_handler     USART1_TX_IRQHandler
  def_irq_handler     LESENSE_IRQHandler
  def_irq_handler     USART2_RX_IRQHandler
  def_irq_handler     USART2_TX_IRQHandler
  def_irq_handler     UART0_RX_IRQHandler
  def_irq_handler     UART0_TX_IRQHandler
  def_irq_handler     UART1_RX_IRQHandler
  def_irq_handler     UART1_TX_IRQHandler
  def_irq_handler     LEUART0_IRQHandler
  def_irq_handler     LEUART1_IRQHandler
  def_irq_handler     LETIMER0_IRQHandler
  def_irq_handler     PCNT0_IRQHandler
  def_irq_handler     PCNT1_IRQHandler
  def_irq_handler     PCNT2_IRQHandler
  def_irq_handler     RTC_IRQHandler
  def_irq_handler     BURTC_IRQHandler
  def_irq_handler     CMU_IRQHandler
  def_irq_handler     VCMP_IRQHandler
  def_irq_handler     LCD_IRQHandler
  def_irq_handler     MSC_IRQHandler
  def_irq_handler     AES_IRQHandler
  def_irq_handler     EBI_IRQHandler
  def_irq_handler     EMU_IRQHandler

.end
